package mobicomp.net;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Vector;

import mobicomp.interfaces.MHMessage;
import mobicomp.interfaces.MHSocket;
import mobicomp.interfaces.Message;
import mobicomp.interfaces.SHSocket;
import mobicomp.util.Dispatcher;

/**
 * TODO
 * 
 * @author Andreas Wetzel <andi@spamato.net>
 */
public class MHSocketImpl implements MHSocket
{
    public static final int DEFAULT_TIMEOUT = 1000;
    public static final int MAX_TTL = 10;
    public static final int MAX_RETRIES= 5;
    
    // our lower layered singlehop socket
    private SHSocket shLayer;
    
    private HashMap routingTable = new HashMap();
    
    // id of the last flood packet
    private int lastMsgId = 0;
    // storage for all flood messages we have seen before
    private HashSet seenFloodMessages = new HashSet();
    
    private Thread messageReceiverThread;
    
    private Dispatcher dispatcher = new Dispatcher(); 
    
    // holds ids of acks we have received but not processed yet
    private Vector receivedAcks = new Vector();


    public MHSocketImpl( int address, MessageListener listener) throws IOException {
        shLayer = new SHSocketImpl( address);
        
        messageReceiverThread = new MHMessageReceiver();
        messageReceiverThread.setDaemon( true);
        messageReceiverThread.start();
        
        dispatcher.addObserver( new RouteRequestHandler( this));
        dispatcher.addObserver( new RouteReplyHandler( this));
        dispatcher.addObserver( new MHMessageHandler( this, listener));
        dispatcher.addObserver( new MHAckHandler( this));
    }
    
    /**
     * @see mobicomp.interfaces.MHSocket#setOwnAddress(int)
     */
    public void setOwnAddress( int addr) {
        shLayer.setOwnAddress( addr);
    }

    /**
     * @see mobicomp.interfaces.MHSocket#getOwnAddress()
     */
    public int getOwnAddress() {
        return shLayer.getOwnAddress();
    }

    /**
     * @see mobicomp.interfaces.MHSocket#send(mobicomp.interfaces.MHMessage)
     */
    public void send( MHMessage message) throws IOException {
		// generate an id
		message.setId( nextMessageId());

		if (message.getReceiver() == BROADCAST_ADDRESS) {
    			broadcastMHMessage( message);
    		}
    		else {
    			// handle addressed messages

    			// try at most MAX_RETRIES times
    			boolean acked = false;
    			for ( int tryy=1; ((!acked) && (tryy<=MAX_RETRIES)); tryy++ ) {
    				// ensure that we know a route to the receiver
    				
    				Vector route = findRoute( message.getReceiver(), (tryy > 1));
//    				System.out.println( "using route " + route);
    		        	message.setRoute( route);
    		        	message.setRouteIndex( 1);
    		        	message.setType( MHMessage.TYPE_SRMSG);
    		        	int nexthop = ((Integer)route.get( 1)).intValue();
	        		sendMHMessage( message, nexthop);

	        		// wait some time for an ack
    		        	acked = waitForAck( message);
    		        	if (acked) {
    		        		break;
    		        	}
 			}
    			
    			// if there was no ack, the destination was unreachabe
    			if (!acked) {
    				throw new IOException( "destination not reachable");
    			}
    		}
    }
    
    private Vector findRoute( int receiver, boolean newRoute) throws IOException {
    		Object receiverHash = new Integer( receiver);
    		if (newRoute) {
    			// delete old route
    			routingTable.remove( receiverHash);
    		}
        for ( int ttl=1; (!routingTable.containsKey( receiverHash) && (ttl <= MAX_TTL)); ttl*=2) {

            // broadcast for new route - assemble route request message
            MHMessage rreq = new MHMessageImpl();
            rreq.setType( MHMessage.TYPE_RREQ);
            rreq.setSender( getOwnAddress());
            // we use the receiver field for the address we're looking for
            rreq.setReceiver( receiver);
            rreq.getRoute().add( new Integer( getOwnAddress()));

            rreq.setTTL( ttl);
            rreq.setId( nextMessageId());
//            System.out.println( "flooding with ttl=" + ttl);
            broadcastMHMessage( rreq);

            synchronized (routingTable) {
                try {
                    routingTable.wait( DEFAULT_TIMEOUT);
                }
                catch (InterruptedException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
        }
        
        Vector route = (Vector)routingTable.get( receiverHash);

        // if we still have no route to host, throw exception
        if (route == null) {
        		throw new IOException( "Cannot route to target " + receiver);
        }
        
        return route;
    }
    
    private boolean waitForAck( MHMessage message) {
    		long started = System.currentTimeMillis();
    		Object hash = calcMsgHash( message);
    		synchronized( receivedAcks) {
    			while ( !receivedAcks.contains( hash) ) {
    				long towait = started + DEFAULT_TIMEOUT - System.currentTimeMillis();
    				if (towait <= 0) {
    					return false;
    				}
    				try {
					receivedAcks.wait( towait);
				}
    				catch (InterruptedException e) {
					e.printStackTrace();
				}
    			}
    			receivedAcks.remove( hash);
//    			System.out.println( getOwnAddress() + ": got ack");
        		return true;
    		}
    }
    
    void sendMHMessage( MHMessage message, int to) throws IOException {
        Message shm = new SHMessageImpl();
        shm.setSender( getOwnAddress());
        shm.setReceiver( to);
        // TODO: this cast is not nice. 
        shm.setData( ((MHMessageImpl)message).toByteArray());
        shLayer.send( shm);
    }

    void broadcastMHMessage( MHMessage message) throws IOException {
        seenFloodMessages.add( calcMsgHash( message));
        sendMHMessage( message, SHSocket.BROADCAST_ADDRESS);
    }

    private int nextMessageId() {
        lastMsgId++;
        if ( lastMsgId > 0xFFFF ) {
            lastMsgId = 1;
        }
        return lastMsgId;
    }
    
    private Object calcMsgHash( int sender, int id) {
    		return new Integer( (sender << 19) + id);
    }
    
    // generate a unique key object for the given MHMessage
    private Object calcMsgHash( MHMessage message) {
        return calcMsgHash( message.getSender(),  message.getId());
    }
    
    void newRoute( int target, Vector route) {
    		synchronized( routingTable) {
    			routingTable.put( new Integer( target), route);
    			routingTable.notifyAll();
    		}
//		System.out.println( getOwnAddress() + ": received route to " + target);
//		System.out.println( "found route: "+route);
    }
    
    void newAck( MHMessage msg) {
    		synchronized( receivedAcks) {
    			receivedAcks.add( calcMsgHash( msg.getReceiver(), msg.getId()));
    			receivedAcks.notifyAll();
    		}
    }
    
    private class MHMessageReceiver extends Thread {
        
        public MHMessageReceiver() {
            
        }
        
        public void run() {
            while (true) {
                try {
                    Message received = shLayer.receive();
//                    System.out.println( getOwnAddress() +": mhsocket got shmessage from shlayer");
                    MHMessage mhm = new MHMessageImpl( received.getData());

                    // if floodId is not zero, it is a broadcast message
                    if ((mhm.getId() != 0) && (mhm.getType() != MHMessage.TYPE_SRACK)) {
                        // check if message already seen
                        Object hash = calcMsgHash( mhm);
                        if (seenFloodMessages.contains( hash)) {
                            // if yes, do nothing
                            continue;
                        }
                        else {
                            // otherwise add to seen flood messages and process message
                            seenFloodMessages.add( hash);
                        }
                    }
                    dispatcher.dispatch( mhm);
                }
                catch (IOException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }                
            }
        }
    }
}
